import pyvista as pv
import numpy as np
import trimesh

from provis.src.processing.surface_handler import SurfaceHandler
from provis.src.processing.name_checker import NameChecker

class Surface:
    """
    The Surface class is used to visualize the surface information of the given molecule.
    
    The class can visualize two kinds of surfaces:
    - the "correct" surface created by the MSMS binary. 
    - a good approximation of the surface computed natively with o3d and trimesh. (MSMS does not have to be installed for this option. It is slightly slower and less precise)
    
    Choose between the two by setting the msms Boolean variable. (True corresponding to the MSMS binary option is default)
    
    """
    def __init__(self, nc, sh=None, msms=False, density=3.0, notebook=False):
        """
        In this constructor no mesh information is computed, simply preparations are made. Meshes are loaded in the plotting function.
        
        A NameChecker object is required for initialization, as this is how the program finds the desired pdb file/molecule.
        If nothing else is passed the NameChecker object will be used to initialize the other internal objects of the surface class.
        
        Apart from the required NameChecker object one can also pass a SurfaceHandler for even more control.
        
        :param name: nc - Instance of a NameChecker class. Used to pass the pdb file name and paths.
        :param type: NameChecker
        :param name: sh - Instance of SurfaceHandler. This class computes the required surface meshes for the plots. The "brain" of the Surface class. Default: None - a SurfaceHandler class will be initialized with the NameChecker object.
        :param type: SurfaceHandler, optional
        :param name: msms - If True plot msms binary version of surface. If False plot the native (non-binary) surface. Default: False.
        :param type: bool, optional
        :param name: density - sampling density used in msms binary. Also needed to load the face and vert files, as their (file)names include the density
        :param type: float, optional
        :param name: notebook - Needs to be set to true to work in a notebook environment. Defualts to False.
        :param type: bool, optional 
        """
        
        self._path, self._out_path, self._base_path = nc.return_all()
        if not sh:
            self._sh = SurfaceHandler(nc, density=density)
        else:
            self._sh = sh
        self._msms = msms
        self._notebook = notebook
        self._shading = not self._notebook

    def plot(self, feature=None, title="Surface", patch=False, box=None, res=None, outname=None, camera=None):
        """
        Plot the surface of the protein.
        
        :param name: feature - Pass which feature (coloring) you want to plot. Options: hydrophob, shape, charge. Default: None (uniform coloring).
        :param type: str, optional
        :param name: title - Title of the plot window. Default: Surface.
        :param type: str, optional
        :param name: patch - If True then coloring will be read in from "root directory"/data/tmp/{pdb_id}.pth file. Default: False.
        :param type: bool, optional
        :param name: box, optional - If True bounding box also visualized, default: 0.
        :param type: bool, optional
        :param name: res - Residues passed in 'res' will be plotted with a bounding box around them. Defaults to None.
        :param type: Residue, optional
        :param name: outname - save image of plot to specified filename. Will appear in data/img/ directory. Default: data/img/{self._out_path}_surface.
        :param type: string, optional
        :param name: camera - Pass a Pyvista Camera https://docs.pyvista.org/api/core/camera.html to manually set the camera position. If nothing/None is passed then the camera position will be set to 'xy'. Default: None.
        :param type: pyvista.Camera, optional
        
        :return: Pyvista.Plotter window - Window with interactive plot.
        """
        # get appropriate mesh and coloring
        mesh, cas = self._sh.return_mesh_and_color(self._msms, feature=feature, patch=patch)
        
        # plot
        pl = pv.Plotter(notebook=self._notebook)
        pl.background_color = 'grey'
        pl.enable_3_lights()
        pl.add_mesh(mesh, scalars=cas, cmap='RdBu', smooth_shading=self._shading, show_edges=False)
        
        # if specified add bounding box
        if box:
            pl.add_bounding_box(color='white', corner_factor=0.5, line_width=1)
        
        if res:
            res_list, chain_list, pad = res.get_res_info()
            for i, r in enumerate(res_list):
                chain = chain_list[i]
                residues = self._dh.get_structure().get_residues()
                residues_list = list(residues)
                res_name = residues_list[r + 1].get_resname()
                d = (self._dh._res_size_dict[res_name] + pad) * 2
                x, y, z = d,d,d
                pl.add_mesh(pv.Cube(center=self._dh.get_residue_info(r, chain,'com'), x_length=x, y_length=y, z_length=z), style='wireframe', show_edges=1, line_width=5, smooth_shading=self._shading, color='r')

        
        # save a screenshot
        if not outname:
            new_name = self._out_path.split('/')
            new_name = new_name[-1].split('.')[0]
            outname = self._base_path + 'data/img/' + new_name + '_surface.png'
            
        if camera: 
            pl.camera = camera
        else:
            pl.camera_position = 'xy'
        pl.show(screenshot=outname, title=title)

    def plot_hydrophob(self, box=None, res=None, outname=None, camera=None):
        """
        Plot the hydrophobic features of a protein.

        :param name: box, optional - If True bounding box also visualized, default: 0.
        :param type: bool, optional
        :param name: res - Residues passed in 'res' will be plotted with a bounding box around them. Defaults to None.
        :param type: Residue, optional
        :param name: outname - save image of plot to specified filename. Will appear in data/img/ directory. Default: data/img/{self._out_path}_surface.
        :param type: string, optional
        :param name: camera - Pass a Pyvista Camera https://docs.pyvista.org/api/core/camera.html to manually set the camera position. If nothing/None is passed then the camera position will be set to 'xy'. Default: None.
        :param type: pyvista.Camera, optional
        
        :return: Pyvista.Plotter window - Window with interactive plot.
        """

        # save a screenshot
        if not outname:
            new_name = self._out_path.split('/')
            new_name = new_name[-1].split('.')[0]
            outname = self._base_path + 'data/img/' + new_name + '_hydrophob.png'
        self.plot(feature="hydrophob", title="Hydrophob", box=box, res=res, outname=outname, camera=camera)

    def plot_shape(self, box=None, res=None, outname=None, camera=None):
        """
        Plot the shape features of a protein.

        :param name: box, optional - If True bounding box also visualized, default: 0.
        :param type: bool, optional
        :param name: res - Residues passed in 'res' will be plotted with a bounding box around them. Defaults to None.
        :param type: Residue, optional
        :param name: outname - save image of plot to specified filename. Will appear in data/img/ directory. Default: data/img/{self._out_path}_surface.
        :param type: string, optional
        :param name: camera - Pass a Pyvista Camera https://docs.pyvista.org/api/core/camera.html to manually set the camera position. If nothing/None is passed then the camera position will be set to 'xy'. Default: None.
        :param type: pyvista.Camera, optional
        
        :return: Pyvista.Plotter window - Window with interactive plot.
        """
        
        # save a screenshot
        if not outname:
            new_name = self._out_path.split('/')
            new_name = new_name[-1].split('.')[0]
            outname = self._base_path + 'data/img/' + new_name + '_shape.png'
        self.plot(feature="shape", title="Shape", box=box, res=res, outname=outname, camera=camera)

    def plot_charge(self, box=None, res=None, outname=None, camera=None):
        """
        Plot the charge features of a protein.

        :param name: box, optional - If True bounding box also visualized, default: 0.
        :param type: bool, optional
        :param name: res - Residues passed in 'res' will be plotted with a bounding box around them. Defaults to None.
        :param type: Residue, optional
        :param name: outname - save image of plot to specified filename. Will appear in data/img/ directory. Default: data/img/{self._out_path}_surface.
        :param type: string, optional
        :param name: camera - Pass a Pyvista Camera https://docs.pyvista.org/api/core/camera.html to manually set the camera position. If nothing/None is passed then the camera position will be set to 'xy'. Default: None.
        :param type: pyvista.Camera, optional
        
        :return: Pyvista.Plotter window - Window with interactive plot.
        """
    
        # save a screenshot
        if not outname:
            new_name = self._out_path.split('/')
            new_name = new_name[-1].split('.')[0]
            outname = self._base_path + 'data/img/' + new_name + '_charge.png'
        self.plot(feature="charge", title="Charge", box=box, res=res, outname=outname, camera=camera)


import pyvista as pv

from provis.src.processing.data_handler import DataHandler
from provis.src.processing.name_checker import NameChecker

class Structure:
    """
    The Structure class is used to visualize the structural information of the given molecule. One can easily plot the atoms, residues, bonds or any combination of these structures.
    """
    def __init__(self, nc, dh=None, plot_solvent=False, notebook=False):
        """
        The constructor creates the internal data structures and loads all the atomic information required for plotting.
        
        A NameChecker object is required for initialization, as this is how the program finds the desired pdb file/molecule.
        If nothing else is passed the NameChecker object will be used to initialize the other internal objects of the structure class.
        
        Apart from the required NameChecker object one can also pass a DataHandler for even more control.
        
        :param name: nc - Instance of a NameChecker class. Used to pass the pdb file name and paths.
        :param type: NameChecker
        :param name: dh - Instance of a DataHandler class. Used to retrieve atom-positional information. Default: None. If None a new DataHandler variable will be initialized with "nc".
        :param type: DataHandler, optional
        :param name: plot_solvent - If True solvent molecules will also be plotted. Default: False.
        :param type: bool, optional
        :param name: notebook - Needs to be set to true to work in a notebook environment. Defualts to False.
        :param type: bool, optional 
        """
        self._notebook = notebook
        self._shading = not self._notebook

        self._id, self._name, self._base_path = nc.return_all()
        # create "brain" of plotting class
        if not dh:
            self._dh = DataHandler(nc)
        else:
            self._dh = dh
        atom_data = self._dh.get_atoms(show_solvent=plot_solvent) # second arg: 1 = show_solvent
        ## return list of spheres (meshes) and colors for the spheres
        self._atoms, self._col_a = self._dh.get_atom_mesh(atom_data, vw=0) # second arg: 1 = showvw spheres instead of "normal" radius
        ## return list of lines (meshes)
        self._bonds, self._bond_col = self._dh.get_bond_mesh()
        ## return list of spheres (meshes) and colors for the spheres
        res_data = self._dh.get_residues()
        self._residues, self._col_r = self._dh.get_residue_mesh(res_data)
        
        self._atoms_vw, self._col_vw = self._dh.get_atom_mesh(atom_data, vw=1)
        

    def manual_plot(self, box=0, res=0, outname=0, atoms=0, col_a=0, bonds=0, vw=0, residues=0, col_r=0, bb=0, camera=None):
        """
        Plot stick and point model. In this function one can pass all the desired meshes to be plotted. One can get these meshes from the DataHandler class.
        
        :param name: box - If True bounding box also visualized, default: 0.
        :param type: bool, optional
        :param name: res - List of pyvista Shperes representing each residue, default: 0.
        :param type: list, optional
        :param name: outname - save image of plot to specified filename. Will appear in data/img directory. default: data/img/{self._name}_stick_point.
        :param type: string, optional
        :param name: atoms - List of pyvista Shperes representing each atom, default: 0.
        :param type: list, optional
        :param name: col_a - List of colors for each atom, default: 0.
        :param type: list, optional
        :param name: bonds - List of pyvista Lines representing each bond, default: 0.
        :param type: list, optional
        :param name: vw - If True styling for Van-der-Waals plotting set. Vw atomic objects still have to be passed under 'atoms' variable.
        :param type: bool, optional
        :param name: col_r - List of colors for each residue, default: 0.
        :param type: list, optional
        :param name: res - Specified residues will be plotted with a bounding box around them.
        :param type: Residue, optional
        :param name: bb - List of coordinates describing the back-bone of the protein, default: 0.
        :param type: bool, optional
        :param name: camera - Pass a Pyvista Camera https://docs.pyvista.org/api/core/camera.html to manually set the camera position. If nothing/None is passed then the camera position will be set to 'xy'. Default: None.
        :param type: pyvista.Camera, optional
        
        :return: Pyvista.Plotter window - Window with interactive plot
        """
        
        pl = pv.Plotter(notebook=self._notebook)
        pl.background_color = 'grey'
        pl.enable_3_lights()

        # adding the spheres (by atom type) one at a time
        opacity = 1 - vw*0.4
        style = 'surface'
        if atoms != 0:
            if vw:
                style='wireframe'
            for j, mesh in enumerate(atoms):
                pl.add_mesh(mesh, color=col_a[j], opacity=opacity, smooth_shading=self._shading, style=style)

        # adding the bonds one at a time
        if bonds != 0:
            for b in bonds:
                pl.add_mesh(b, color='w', render_lines_as_tubes=True, line_width=5)
        # adding the spheres (by residue) one at a time
        # only executes if residue information provided
        if residues != 0:
            for k, mesh in enumerate(residues):
                pl.add_mesh(mesh, color=col_r[k], opacity=0.2)
        
        # if specified add bounding box
        if box:
            pl.add_bounding_box(color='white', corner_factor=0.5, line_width=1)
        if res:
            res_list, chain_list, pad = res.get_res_info()
            for i, r in enumerate(res_list):
                chain = chain_list[i]
                residues = self._dh.get_structure().get_residues()
                residues_list = list(residues)
                res_name = residues_list[r + 1].get_resname()
                d = (self._dh._res_size_dict[res_name] + pad) * 2
                x, y, z = d,d,d
                pl.add_mesh(pv.Cube(center=self._dh.get_residue_info(r, chain,'com'), x_length=x, y_length=y, z_length=z), style='wireframe', show_edges=1, line_width=5, color='r')
        
        
        if bb:
            pl.add_mesh(pv.Spline(bb), render_lines_as_tubes=True, smooth_shading=self._shading, line_width=10)
        
        # save a screenshot
        if not outname:
            new_name = self._name.split('/')
            new_name = new_name[-1].split('.')[0]
            outname = self._base_path + 'data/img/' + new_name + '_stick_point.png'
        
        if camera: 
            pl.camera = camera
        else:
            pl.camera.position = 'xy'
        pl.show(screenshot=outname, title='Provis')


    def plot(self, box=0, res=None, outname=0, atoms=0, bonds=0, vw=0, residues=0, bb=0, title=None, camera=None):
        """
        This member function is called by all the others. Using this function you can plot any combination of the results gotten from the specialized member functions. For example you could plot the atoms and the backbone of the protein in the same plot.
        
        All information to be plotted is already computed. This function simply dictates what is to be plotted.
        
        :param name: box, optional - If True bounding box also visualized, default: 0.
        :param type: bool, optional
        :param name: res - Residues passed in 'res' will be plotted with a bounding box around them. Defaults to None.
        :param type: Residue, optional
        :param name: outname - Save image of plot to specified filename. Will appear in data/img directory. Defaults to data/img/{pdb_id}_stick_point.png.
        :param type: string, optional
        :param name: atoms - Plot atoms, default: 0.
        :param type: bool, optional
        :param name: bonds, optional - Plot bond. If zero or undefined then it does not plot the bonds, if 1 it plots all bonds uniformly, if 2 it plots colorful bonds (see data_handler). Default: 0.
        :param type: int, optional
        :param name: vw - Plot Wan-der-Waals radii instead of atomic radii.
        :param type: bool, optional
        :param name: residues, optional - Plot residue, default: 0.
        :param type: bool, optional
        :param name: bb - If True backbone of protein is plotted. Default: False.
        :param type: bool, optional
        :param name: title - Title of the plot window. Defaults to None.
        :param type: str, optional
        :param name: camera - Pass a Pyvista Camera https://docs.pyvista.org/api/core/camera.html to manually set the camera position. If nothing/None is passed then the camera position will be set to 'xy'. Default: None.
        :param type: pyvista.Camera, optional
        
        :return: Pyvista.Plotter window - Window with interactive plot
        """
        pl = pv.Plotter(notebook=self._notebook)
        pl.background_color = 'grey'
        pl.enable_3_lights()

        # adding the spheres (by atom type) one at a time
        opacity = 1 - vw*0.4
        style = 'surface'
        if atoms:
            if vw:
                style='wireframe'
                for j, mesh in enumerate(self._atoms_vw):
                    pl.add_mesh(mesh, color=self._col_vw[j], opacity=opacity, smooth_shading=self._shading, style=style)
            else:
                for j, mesh in enumerate(self._atoms):
                    pl.add_mesh(mesh, color=self._col_a[j], opacity=opacity, smooth_shading=self._shading, style=style)

        # adding the bonds one at a time
        if bonds == 1:
            for b in self._bonds:
                pl.add_mesh(b, color="w", line_width=5, render_lines_as_tubes=True)
        if bonds == 2:
            for b, c in zip(self._bonds, self._bond_col):
                pl.add_mesh(b, color=c, line_width=5, render_lines_as_tubes=True)
                
                
        # adding the spheres (by residue) one at a time
        # only executes if residue information provided
        if residues:
            for k, mesh in enumerate(self._residues):
                pl.add_mesh(mesh, color=self._col_r[k], opacity=0.2)
        
        # if specified add bounding box
        if box:
            pl.add_bounding_box(color='white', corner_factor=0.5, line_width=1)
        
        if res:
            res_list, chain_list, pad = res.get_res_info()
            for i, r in enumerate(res_list):
                chain = chain_list[i]
                residues = self._dh.get_structure().get_residues()
                residues_list = list(residues)
                res_name = residues_list[r + 1].get_resname()
                d = (self._dh._res_size_dict[res_name] + pad) * 2
                x, y, z = d,d,d
                pl.add_mesh(pv.Cube(center=self._dh.get_residue_info(r, chain,'com'), x_length=x, y_length=y, z_length=z), style='wireframe', show_edges=1, line_width=5, smooth_shading=self._shading, color='r')

        if bb:
            bb_mesh = self._dh.get_backbone_mesh()
            pl.add_mesh(pv.Spline(bb_mesh), render_lines_as_tubes=True, line_width=10)
        # save a screenshot
        if not outname:
            new_name = self._name.split('/')
            new_name = new_name[-1].split('.')[0]
            outname = self._base_path + 'data/img/' + new_name + '_stick_point.png'
           
        if camera: 
            pl.camera.position = camera
        pl.show(screenshot=outname, title=title)


    def plot_stick_point(self, box=0, res=None, outname=0, camera=None):
        """
        Plot stick and point model of the protein. Atoms are spheres, bonds are tubes.
        
        :param name: box, optional - If True bounding box also visualized, default: 0.
        :param type: bool, optional
        :param name: res - Residues passed in 'res' will be plotted with a bounding box around them. Defaults to None.
        :param type: Residue, optional
        :param name: outname - Save image of plot to specified filename. Will appear in data/img directory. Defaults to data/img/{pdb_id}_stick_point.png.
        :param type: string, optional
        :param name: camera - Pass a Pyvista Camera https://docs.pyvista.org/api/core/camera.html to manually set the camera position. If nothing/None is passed then the camera position will be set to 'xy'. Default: None.
        :param type: pyvista.Camera, optional
        
        :return: Pyvista.Plotter window - Window with interactive plot.
        """

        self.plot(atoms=1, box=box, vw=0, bonds=1, residues=0, res=res, outname=outname, title="Stick Point", camera=camera)
        
    def plot_atoms(self, box=0, res=0, outname=0, camera=None):
        """
        Plot the atoms as spheres.
        
        Consult https://en.wikipedia.org/wiki/CPK_coloring for the coloring. 
        
        :param name: box, optional - If True bounding box also visualized, default: 0.
        :param type: bool, optional
        :param name: res - Residues passed in 'res' will be plotted with a bounding box around them. Defaults to None.
        :param type: Residue, optional
        :param name: outname - Save image of plot to specified filename. Will appear in data/img directory. Defaults to data/img/{pdb_id}_atoms.png.
        :param type: string, optional
        :param name: camera - Pass a Pyvista Camera https://docs.pyvista.org/api/core/camera.html to manually set the camera position. If nothing/None is passed then the camera position will be set to 'xy'. Default: None.
        :param type: pyvista.Camera, optional
        
        :return: Pyvista.Plotter window - Window with interactive plot.
        """
        # save a screenshot
        if not outname:
            new_name = self._name.split('/')
            new_name = new_name[-1].split('.')[0]
            outname = self._base_path + 'data/img/' + new_name + '_atoms.png'
        self.plot(atoms=1, box=box, vw=0, bonds=0, residues=0, res=res, outname=outname, title="Atoms", camera=camera)
        
    def plot_vw(self, box=0, res=0, outname=0, camera=None):
        """
        Plot Van-der-Waals radius of atoms as wireframe spheres.
        
        :param name: box, optional - If True bounding box also visualized, default: 0.
        :param type: bool, optional
        :param name: res - Residues passed in 'res' will be plotted with a bounding box around them. Defaults to None.
        :param type: Residue, optional
        :param name: outname - Save image of plot to specified filename. Will appear in data/img directory. Defaults to data/img/{pdb_id}_atoms.png.
        :param type: string, optional
        :param name: camera - Pass a Pyvista Camera https://docs.pyvista.org/api/core/camera.html to manually set the camera position. If nothing/None is passed then the camera position will be set to 'xy'. Default: None.
        :param type: pyvista.Camera, optional
        
        :return: Pyvista.Plotter window - Window with interactive plot.
        """
        # save a screenshot
        if not outname:
            new_name = self._name.split('/')
            new_name = new_name[-1].split('.')[0]
            outname = self._base_path + 'data/img/' + new_name + '_vw.png'
        self.plot(atoms=1, box=box, vw=1, bonds=0, residues=0, res=res, title="Van-der-Waals", camera=camera)
        
    def plot_bonds(self, box=0, res=0, outname=0, colorful=False, camera=None):
        """
        Plot only the bonds. By default all bonds will be plotted uniformly. If you want to view the difference in bonds you can set the colorful variable to True.
        
        Single bonds: white
        Double bonds: blue
        Triple bonds: green
        Amide bonds: red
        Aromatic bonds: purple
        Undefined/Anything else: black
        
        :param name: box - If True bounding box also visualized, default: 0.
        :param type: bool, optional
        :param name: res - Residues passed in 'res' will be plotted with a bounding box around them. Defaults to None.
        :param type: Residue, optional
        :param name: outname - Save image of plot to specified filename. Will appear in data/img directory. Defaults to data/img/{pdb_id}_atoms.png.
        :param type: string, optional
        :param name: colorful - If True bonds will be plotted in a colorful manner. If False all bonds are white. Default: False
        :param type: bool, optional
        :param name: camera - Pass a Pyvista Camera https://docs.pyvista.org/api/core/camera.html to manually set the camera position. If nothing/None is passed then the camera position will be set to 'xy'. Default: None.
        :param type: pyvista.Camera, optional
        
        :return: Pyvista.Plotter window - Window with interactive plot.
        """
        b = bool(colorful) * 1 + 1
        # save a screenshot
        if not outname:
            new_name = self._name.split('/')
            new_name = new_name[-1].split('.')[0]
            outname = self._base_path + 'data/img/' + new_name + '_bonds.png'
        self.plot(atoms=0, box=box, vw=0, bonds=b, residues=0, res=res, outname=outname, title="Bonds", camera=camera)
        
        
    def plot_backbone(self, box=0, res=0, outname=0, camera=None):
        """
        Plots the backbone (roughly the amide bonds) of the protein.
        
        :param name: box - If True bounding box also visualized, default: 0.
        :param type: bool, optional
        :param name: res - Residues passed in 'res' will be plotted with a bounding box around them. Defaults to None.
        :param type: Residue, optional
        :param name: outname - Save image of plot to specified filename. Will appear in data/img directory. Defaults to data/img/{pdb_id}_backbone.png.
        :param type: string, optional
        :param name: camera - Pass a Pyvista Camera https://docs.pyvista.org/api/core/camera.html to manually set the camera position. If nothing/None is passed then the camera position will be set to 'xy'. Default: None.
        :param type: pyvista.Camera, optional
        
        :return: Pyvista.Plotter window - Window with interactive plot.
        """
        # save a screenshot
        if not outname:
            new_name = self._name.split('/')
            new_name = new_name[-1].split('.')[0]
            outname = self._base_path + 'data/img/' + new_name + '_backbone.png'
        self.plot(atoms=0, box=box, vw=0, bonds=0, residues=0, res=res, bb=1, outname=outname, title="Backbone", camera=camera)




from genericpath import exists
import os
import provis

class NameChecker:
    """
    The point of this class is to provide uniform naming and path locations to all the other classes of provis.
    NameChecker has internal variables and a method to return these variables. 
    
    Variables: 
    self._pdb_name - Full path to the pdb file without the .pdb extension. Usually PROVIS_PATH/data/pdb/{pdb_id}.
    self._out_path - Full path to the temporary files. The names of all temporary files are derived from this variable. Usually PROVIS_PATH/data/tmp/{pdb_id}.
    self._base_path - Full path of the provis directory or any directory that has the following directory structure within: {path}/data/data, {path}/data/img, {path}/data/tmp, {path}/data/pdb. 
    """
    
    def __init__(self, name, base_path: str=None):
        """
        Checks if the provided name corresponds to an existing pdb file.
        The "name" variable is a path to the pdb file with or without the pdb extension. One can also simply pass the pdb ID (name without extension) if the file is saved under: "root directory"/data/pdb.
        Always sets "out_path" to default output location: "root directory"/data/tmp.
        
        This class is also a global class and ensures that all the plotting classes plot the same pdb file. The class has to be initialized once, after that everything simply uses the variables from the global class.

        :param name: name - Name and path to location of the pdb file. If simply pdb id (file name without .pdb extension) is provided the default directory - provis/data/pdb - will be searched for the file. If no file found the input is returned.
        :param type: str
        :param name: base_path - Easiest if you leave this alone, clone the provis git repo and run everything in the provis root directory. Path to the base directory with the following file structure: {base_path}/data/data, {base_path}/data/img, {base_path}/data/tmp, {base_path}/data/pdb AND {base_path}/binaries that contains the binaries required to run provis. Defaults to the "provis" packages path.
        :param type: str, optional
        """
        if not base_path:
            path = os.getcwd() #path.dirname(provis.__file__)
            # path points to provis/provis, we only want provis/
            base_path = path #[0:-6]
            if base_path.split("/")[-1] == "examples":
                base_path = base_path[0:-8]
            
        if base_path[-1] != "/":
            base_path += "/"
        
        name = name.split(".")[0]
        name_pdb = name + ".pdb"

        unkown_name = name.split("/")[-1]
        id = unkown_name.split(".")[0]

        
        if not (exists(name_pdb)):
            # name = name.split(".")[0]
            default = base_path + "data/pdb/" + id
            default_file = default + ".pdb"
            if exists(default_file):
                name = default
            else:
                print("The provided pdb file could not be found. Please make sure you have it downloaded and provided the correct file path.")


        out_path = base_path + "data/tmp/" 
        out_path += id 
        self._pdb_name = name 
        self._out_path = out_path
        self._base_path = base_path
    
    def return_all(self):
        """
        Return all class variables.
        
        :return: str - path to the input pdb id (file without the extension)
        :return: str - path to the output pdb id - data/tmp/{pdb_id}
        :return: str - path to the root of the provis directory. Following file structure HAS TO exist within: {path}/data/data, {path}/data/img, {path}/data/tmp, {path}/data/pdb.
        """
        
        return self._pdb_name, self._out_path, self._base_path




# imports
from genericpath import sameopenfile
import os

import numpy as np
from numpy.lib.twodim_base import tri
import torch
import pyvista as pv
from pyvtk import PolyData, PointData, CellData, Scalars, Vectors, VtkData
import trimesh
from subprocess import PIPE, Popen
import open3d as o3d
import enum
from os.path import exists

from provis.utils.surface_utils import get_surface, compute_normal, prepare_trimesh, fix_trimesh, find_nearest_atom
from provis.utils.surface_feat import compute_surface_features
from provis.src.processing.file_converter import FileConverter
from provis.src.processing.data_handler import DataHandler

class SurfaceHandler:
    """
    The 'brain' of provis, when it comes to handling surface information. 
    
    This class loads information from a variety of files and creates meshes to be plotted. 
    Upper level classes - eg. StickPoint - have their own AtomHandler objects that do all the work.    
    """
  
    def __init__(self, nc, fc=None, dh=None, density=3.0):
        """
        Initializes SurfaceHandler class memeber variables: 
        _dh: DataHandler - needed for native mesh creation.
        _fc: FileConverter - needed to create necessairy files.   
        _path: str - path to pdb file (without extension).
        _out_path: str - path to output files (without extension).
        _base_path: str - path to root directory.
        _density: float - density of surface mesh.
        _features: tuple(numpy.ndarray) - A collection of arrays representing the feature information for the surface of the protein, saved in a tuple.
        _mesh: Mesh - Surface mesh of protein.
        _col: numpy.ndarray - Coloring map of surface. To be passed to pyvista.PolyData.add_mesh(scalars=).
        _res_id: list - List of residue IDs (format from output_pdb_as_xyzrn())
        _atom_coords: list - List of atomic coordinates. Just the coordinates in the same order as in _res_id.

        :param name: nc - Instance of a NameChecker class. Used to pass the pdb file name and paths.
        :param type: NameChecker
        :param name: fc - Instance of a FileConverter class. Used to create the necessairy files (face, vert, pqr, ...) if they do not exist yet. Default: None.
        :param type: FileConverter
        :param name: dh - Instance of a DataHandler class. Used to retrieve atom-positional information when calculating the surface of the protein natively. Default: None. If None a new DataHandler variable will be initialized with "nc".
        :param type: DataHandler, optional
        :param name: density - Density needed for msms. Default: 3.0.
        :param type: float, optional
        """
        
        if not fc:
            self._fc = FileConverter(nc)
        else:
            self._fc = fc
        if not dh:
            self._dh = DataHandler(nc, fc)
        else:
            self._dh = dh
        self._path, self._out_path, self._base_path = nc.return_all()
        self._density = density
        self._features = None
        self._mesh = None
        self._col = None
        self._res_id = None
        self._atom_coords = None
        
    def get_assignments(self):
        """
        Get assignments (coloring) for the mesh. File has to exist, no way to produce it with provis. 
        Loads "root directory"/data/tmp/{pdb_id}.pth and returns it.

        :returns: PyTourch object - Coloring of surface.
        """
        filename = self._out_path.upper() + '.pth'
        assigment = torch.load(filename)
        
        return assigment

    def get_surface_features(self, mesh, feature,  res_id=None):
        """
        Get the coloring corresponding to a specific feature.

        :param name: mesh - The mesh
        :param type: Trimesh
        :param name: feature - Name of feature we are interested in. Options: hydrophob, shape, charge.
        :param type: str
        :param name: native - Set to True when using native mesh computation. Default: False.
        :param type: bool, optional
        
        :raises: NotImplementedError - If unkown feature specified error is raised

        returns: numpy.ndarray - Array of coloring corresponding to surface.
        """

        # get surface
        pdb_file = self._path + '.pdb'
        
        pqr_file = self._out_path + '.pqr'
        if not exists(pqr_file):
            self._fc.pdb_to_pqr(self._path, self._out_path)
        
        if not res_id:
            surface = get_surface(self._out_path, density=self._density)
        else:
            full_res_id = find_nearest_atom(self._atom_coords, res_id, mesh.vertices)
            
            surface = [mesh.vertices, mesh.faces, mesh.face_normals, full_res_id, mesh.area_faces]
            
        # compute features of surface
        if not self._features:
            self._features = compute_surface_features(surface, pdb_file, self._out_path, mesh, pdb_id = self._path)

        if feature == 'hydrophob':
            return self._features[2]
        elif feature == 'shape':
            return self._features[0]
        elif feature == 'charge':
            return self._features[3]
        else:
            raise NotImplementedError

    def msms_mesh_and_color(self, feature=None, patch=False):
        """
        Return the mesh and coloring ready for plotting.

        :param name: feature - Name of feature, same as in get_surface_features. Options: hydrophob, shape, charge. Defaults to None.
        :param type: str, optional
        :param name: patch - Set coloring of mesh manually, from a file. If set to True get_assignments() will be called. Defaults to False.
        :param type: bool, optional

        :returns: Trimesh - The mesh, always returned
        :returns: numpy.ndarray - Color map corresponding to specified feature. Only returned if a feature is specified as an input. This map can be passed to a pyvista.Plotter.add_mesh() function as the 'scalars' argument to get the encoded coloring.
        """
        
        path = f"{self._out_path}_out_{int(self._density * 10)}"
        face = path + '.face'
        vert = path + '.vert'
        from os.path import exists
        # Check if face file exists. If not convert it from pdb
        file_exists = exists(face) and exists(vert)
        if not file_exists:
            self._fc.msms(self._out_path, self._density)
        
        if not self._mesh:
            surface = get_surface(out_path=self._out_path, density=self._density)
            self._mesh = prepare_trimesh(vertices=surface[0], faces=surface[1], normals=surface[2], 
                            resolution=1.5, apply_fixes=True)
        
        if feature:    
            if patch:
                self._col = self.get_assignments()
            if not patch:
                self._col = self.get_surface_features(self._mesh, feature)
        else:
            self._col = None

    def native_mesh_and_color(self, feature=None):
        """
        Returns a mesh without the need for the MSMS binary.
        Always returns a pyvista.PolyData mesh of the surface and if a feature is specified it also returns the coloring according to that feature.

        :param name: feature - Name of feature, same as in get_surface_features. Options: hydrophob, shape, charge. Defaults to "".
        :param type: str, optional

        :returns: pyvista.PolyData - The mesh, always returned
        :returns: numpy.ndarray - Color map corresponding to specified feature. Only returned if a feature is specified as an input. This map can be passed to a pyvista.Plotter.add_mesh() function as the 'scalars' argument to get the encoded coloring.
        """
        if not self._mesh:
            # create rough surface by combining vw radii of each atom
            atom_data, self._res_id, self._atom_coords = self._dh.get_atoms_IDs()
            self._atmsurf, col = self._dh.get_atom_mesh(atom_data, vw=1, probe=0.1)

            # adding the spheres (by atom type) one at a time
            j = 0
            mesh_ = pv.wrap(self._atmsurf[0])
            for mesh in self._atmsurf[1:]:
                mesh_ = mesh_ + (mesh)

            # blur the spheres and extract edges (pyvista)
            mesh_ = mesh_.delaunay_3d(alpha=1.5).extract_feature_edges()
            # create trimesh mesh
            new = trimesh.Trimesh(mesh_.points)
            cloud = o3d.geometry.PointCloud()
            cloud.points = o3d.utility.Vector3dVector(new.vertices)
            cloud.normals = o3d.utility.Vector3dVector(new.vertex_normals)
            # reconstruct the surface mesh as a trimesh mesh
            radii = [0.1, 0.3, 0.45, 0.6, 0.75, 0.9,1.2, 1.5]
            tri_mesh= o3d.geometry.TriangleMesh.create_from_point_cloud_ball_pivoting(cloud, o3d.utility.DoubleVector(radii))#, depth=depth, width=width, scale=scale, linear_fit=linear_fit)
            v = np.asarray(tri_mesh.vertices)
            f = np.array(tri_mesh.triangles)
            f = np.c_[np.full(len(f), 3), f]
            mesh = pv.PolyData(v, f)
            shell =  mesh.clean().reconstruct_surface()#.clean()
            shell.compute_normals(inplace=True)

            # parse list of PolyData (format) faces and convert them into Trimesh (format) faces 
            len_f = len(shell.faces)
            faces_ = []
            i = 0
            while(i < len_f):
                curr = shell.faces[i]
                temp = [None] * curr
                idxx = 0
                j = i + 1
                while(curr):
                    temp[idxx] = shell.faces[j + idxx]
                    idxx += 1
                    curr -= 1
                faces_.append(temp)
                i += idxx + 1
            
            tri_mesh = trimesh.Trimesh(shell.points, faces=faces_)
            self._mesh = tri_mesh

        # only needed if feature is specified (-> color map needs to be calculated)
        if feature:
            self._col = self.get_surface_features(self._mesh, feature, self._res_id)
        else:
            self._col = None

    
    def return_mesh_and_color(self, msms=False, feature=None, patch=False):
        """
        Wrapper function to choose between the msms surface visualization vs the native surface visualization. 
        If you could not download the MSMS binary leave this variable false and you will not run into problems.
        
        :param name: msms - If true, surface generated by msms binary is returned, else the native mesh. Default: False.
        :param type: bool, optional
        :param name: feature - Name of feature, same as in get_surface_features. Options: hydrophob, shape, charge. Defaults to None.
        :param type: str, optional
        :param name: patch - Set coloring of mesh manually. If set to True get_assignments() will be called. Defaults to False.
        :param type: bool, optional
        
        :returns: trimesh.Trimesh - The mesh corresponding to the surface of the protein.
        :returns: numpy.ndarray - Coloring map corresponding to specified feature.
        """
        
        # run appropriate function to compute the mesh and the coloring
        if msms:
            self.msms_mesh_and_color(feature, patch)
        else:
            self.native_mesh_and_color(feature)
          
        return self._mesh, self._col
